@@ -91,6 +91,7 @@ our %EXPORT_TAGS = (
 				       mark_referenced
 				       ensure_chain
 				       ensure_mangle_chain
+				       ensure_nat_chain
 				       new_standard_chain
 				       new_builtin_chain
 				       ensure_filter_chain
@@ -670,6 +671,16 @@ sub ensure_mangle_chain($) {
     $chainref;
 }
 
+sub ensure_nat_chain($) {
+    my $chain = $_[0];
+
+    my $chainref = ensure_chain 'nat', $chain;
+
+    $chainref->{referenced} = 1;
+
+    $chainref;
+}
+
 #
 # Add a builtin chain
 #
@@ -743,7 +754,7 @@ sub finish_chain_section ($$) {
 
     add_rule $chainref, "-m state --state $state -j ACCEPT" unless $config{FASTACCEPT};
 
-    if ($sections{RELATED} ) {
+    if ($sections{NEW} ) {
 	if ( $chainref->{is_policy} ) {
 	    if ( $chainref->{synparams} ) {
 		my $synchainref = ensure_chain 'filter', syn_flood_chain $chainref;
@@ -869,7 +880,7 @@ sub validate_port( $$ ) {
     if ( $port =~ /^(\d+)$/ ) {
 	return $port if $port <= 65535;
     } else {
-	$proto = getprotobyname $proto if $proto =~ /^(\d+)$/;
+	$proto = getprotobynumber $proto if $proto =~ /^(\d+)$/;
 	$value = getservbyname( $port, $proto );
     }
 
@@ -1012,12 +1023,13 @@ sub do_proto( $$$ )
 	  PROTO:
 	    {
 
-		if ( $proto == TCP || $proto == UDP ) {
+		if ( $proto == TCP || $proto == UDP || $proto == SCTP ) {
 		    my $multiport = 0;
 
 		    if ( $ports ne '' ) {
 			if ( $ports =~ tr/,/,/ > 0 || $sports =~ tr/,/,/ > 0 ) {
 			    fatal_error "Port lists require Multiport support in your kernel/iptables" unless $capabilities{MULTIPORT};
+			    fatal_error "Multiple ports not supported with SCTP" if $proto == SCTP;
 			    $ports = validate_port_list $pname , $ports;
 			    $output .= "-m multiport --dports $ports ";
 			    $multiport = 1;
@@ -1124,14 +1136,14 @@ sub do_test ( $$ )
 {
     my ($testval, $mask) = @_;
 
-    return '' unless $testval and $testval ne '-';
+    return '' unless defined $testval and $testval ne '-';
 
     my $invert = $testval =~ s/^!// ? '! ' : '';
     my $match  = $testval =~ s/:C$// ? "-m connmark ${invert}--mark" : "-m mark ${invert}--mark";
 
     validate_mark $testval;
 
-    $testval .= '/0xFF' unless ( $testval =~ '/' );
+    $testval = join( '/', $testval, in_hex( $mask ) ) unless ( $testval =~ '/' );
 
     "$match $testval ";
 }
@@ -1180,8 +1192,9 @@ sub do_user( $ ) {
     } elsif ( $user =~ /^(.*):(.*)$/ ) {
 	$rule .= "--uid-owner $1 " if defined $1 && $1 ne '';
 	$rule .= "--gid-owner $2 " if defined $2 && $2 ne '';
-    } elsif ( $user =~ /^!/ ) {
-	$rule .= "! --uid-owner $user ";
+    } elsif ( $user =~ /^!(.*)$/ ) {
+	fatal_error "Invalid USER/GROUP (!)" if $1 eq '';
+	$rule .= "! --uid-owner $1 ";
     } else {
 	$rule .= "--uid-owner $user ";
     }
@@ -1250,6 +1263,8 @@ sub get_set_flags( $$ ) {
     my ( $setname, $option ) = @_;
     my $options = $option;
 
+    $setname =~ s/^!//;  # Caller has handled !
+
     if ( $setname =~ /^(.*)\[([1-6])\]$/ ) {
 	$setname  = $1;
 	my $count = $2;
@@ -1667,9 +1682,9 @@ sub expand_rule( $$$$$$$$$$ )
 		}
 
 		add_command( $chainref , "for address in $list; do" );
-
 		$rule .= '-d $address ';
 		incr_cmd_level $chainref;
+		add_command( $chainref , '[ $address = 0.0.0.0 ] && continue' );
 	    } else {
 		$rule .= join ( '', '-d ', get_interface_address( $interfaces[0] ), ' ' );
 	    }
@@ -1735,6 +1750,7 @@ sub expand_rule( $$$$$$$$$$ )
 		add_command( $chainref , "for address in $list; do" );
 		$rule .= '-m conntrack --ctorigdst $address ';
 		incr_cmd_level $chainref;
+		add_command( $chainref , '[ $address = 0.0.0.0 ] && continue' );
 	    } else {
 		get_interface_address $interfaces[0];
 		$rule .= join( '', '-m conntrack --ctorigdst $', interface_address ( $interfaces[0] ), ' ' );
@@ -1820,6 +1836,9 @@ sub expand_rule( $$$$$$$$$$ )
     $dnets = ALLIPv4 unless $dnets;
     $onets = ALLIPv4 unless $onets;
 
+    fatal_error "Input interface may not be specified with a source IP address in the POSTROUTING chain"      if $restriction == POSTROUTE_RESTRICT && $iiface && $inets ne ALLIPv4;
+    fatal_error "Output interface may not be specified with a destination IP address in the PREROUTING chain" if $restriction == PREROUTE_RESTRICT &&  $diface && $dnets ne ALLIPv4;
+
     if ( $iexcl || $dexcl || $oexcl ) {
 	#
 	# We have non-trivial exclusion -- need to create an exclusion chain
