@@ -54,6 +54,7 @@ our @EXPORT_OK = qw( $shorewall_dir initialize read_a_line1 set_config_path shor
 
 our %EXPORT_TAGS = ( internal => [ qw( create_temp_object 
 				       finalize_object
+		                       in_hex
 				       emit
 				       emit_unindented
 				       save_progress_message
@@ -102,7 +103,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_object
 
 Exporter::export_ok_tags('internal');
 
-our $VERSION = 4.0.6;
+our $VERSION = 4.0.7;
 
 #
 # describe the current command, it's present progressive, and it's completion.
@@ -243,7 +244,7 @@ sub initialize() {
 		    ORIGINAL_POLICY_MATCH => '',
 		    LOGPARMS => '',
 		    TC_SCRIPT => '',
-		    VERSION =>  '4.0.6',
+		    VERSION => "4.0.12",
 		    CAPVERSION => 40006 ,
 		  );
     #
@@ -338,13 +339,14 @@ sub initialize() {
 		DELETE_THEN_ADD => undef,
 		MULTICAST => undef,
 		DONT_LOAD => '',
+		BROKEN_ROUTING => '',
 		#
 		# Packet Disposition
 		#
 		MACLIST_DISPOSITION => undef,
 		TCP_FLAGS_DISPOSITION => undef,
 		BLACKLIST_DISPOSITION => undef,
-		);
+	      );
 
     #
     # From parsing the capabilities file
@@ -382,7 +384,7 @@ sub initialize() {
 	       HASHLIMIT_MATCH => undef,
 	       NFQUEUE_TARGET => undef,
 	       CAPVERSION => undef,
-	       );
+	     );
     #
     # Directories to search for configuration files
     #
@@ -409,6 +411,13 @@ sub initialize() {
 
 INIT {
     initialize;
+    #
+    # These variables appear within single quotes in shorewall.conf -- add them to ENV
+    # so that read_a_line doesn't have to be smart enough to parse that usage.
+    #
+    for ( qw/root system command files destination/ ) {
+	$ENV{$_} = '' unless exists $ENV{$_};
+    }
 }
 
 #
@@ -447,6 +456,13 @@ sub fatal_error1	{
     die "   ERROR: @_\n";
 }
 
+#
+# Return the argument expressed in Hex
+#
+sub in_hex( $ ) {
+    sprintf '0x%x', $_[0];
+}
+
 #
 # Write the arguments to the object file (if any) with the current indentation.
 #
@@ -731,6 +747,7 @@ sub split_line( $$$ ) {
     my ( $mincolumns, $maxcolumns, $description ) = @_;
 
     fatal_error "Shorewall Configuration file entries may not contain single quotes, double quotes, single back quotes or backslashes" if $currentline =~ /["'`\\]/;
+    fatal_error "Non-ASCII gunk in file" if $currentline =~ /[^\s[:print:]]/;
 
     my @line = split( ' ', $currentline );
 
@@ -755,6 +772,7 @@ sub split_line1( $$$ ) {
     return @line if $line[0] eq 'COMMENT';
 
     fatal_error "Shorewall Configuration file entries may not contain single quotes" if $currentline =~ /'/;
+    fatal_error "Non-ASCII gunk in file" if $currentline =~ /[^\s[:print:]]/;
 
     fatal_error "Invalid $description entry (too few columns)"  if @line < $mincolumns;
     fatal_error "Invalid $description entry (too many columns)" if @line > $maxcolumns;
@@ -785,7 +803,10 @@ sub split_line2( $$$ ) {
     my $columns = $no_pad{$first};
 
     if ( defined $columns ) {
-	fatal_error "Invalid $first entry" if $columns && @line != $columns;
+	if ( $columns ) {
+	    fatal_error "Invalid $first entry" if @line != $columns;
+	}
+
 	return @line
     }
 
@@ -816,7 +837,7 @@ sub open_file( $ ) {
 
     fatal_error 'Internal Error in open_file()' if defined $currentfile;
 
-    do_open_file $fname if -f $fname && -s _;
+    -f $fname && -s _ ? do_open_file $fname : '';
 }
 
 #
@@ -1063,7 +1084,12 @@ sub read_a_line() {
 		#                            $1      $2      $3           -     $4
 		while ( $currentline =~ m( ^(.*?) \$({)? ([a-zA-Z]\w*) (?(2)}) (.*)$ )x ) {
 		    my $val = $ENV{$3};
-		    $val = '' unless defined $val;
+
+		    unless ( defined $val ) {
+			fatal_error "Undefined shell variable (\$$3)" unless exists $ENV{$3};
+			$val = '';
+		    }
+
 		    $currentline = join( '', $1 , $val , $4 );
 		    fatal_error "Variable Expansion Loop" if ++$count > 100;
 		}
@@ -1109,6 +1135,7 @@ sub read_a_line1() {
 	    chomp $currentline;
 	    next if $currentline =~ /^\s*$/;
 	    $currentline =~ s/#.*$//;       # Remove Trailing Comments
+	    fatal_error "Non-ASCII gunk in file" if $currentline =~ /[^\s[:print:]]/;
 	    $currentlinenumber = $.;
 	    return 1;
 	}
@@ -1145,25 +1172,25 @@ sub default_yes_no ( $$ ) {
     }
 }
 
-my %validlevels = ( debug   => 7,
-		    info    => 6,
-		    notice  => 5,
-		    warning => 4,
-		    warn    => 4,
-		    err     => 3,
-		    error   => 3,
-		    crit    => 2,
-		    alert   => 1,
-		    emerg   => 0,
-		    panic   => 0,
-		    none    => '',
+my %validlevels = ( DEBUG   => 7,
+		    INFO    => 6,
+		    NOTICE  => 5,
+		    WARNING => 4,
+		    WARN    => 4,
+		    ERR     => 3,
+		    ERROR   => 3,
+		    CRIT    => 2,
+		    ALERT   => 1,
+		    EMERG   => 0,
+		    PANIC   => 0,
+		    NONE    => '',
 		    ULOG    => 'ULOG' );
 
 #
 # Validate a log level -- Drop the trailing '!' and translate to numeric value if appropriate"
 #
 sub validate_level( $ ) {
-    my $level = $_[0];
+    my $level = uc $_[0];
 
     if ( defined $level && $level ne '' ) {
 	$level =~ s/!$//;
@@ -1321,6 +1348,11 @@ sub qt( $ ) {
     system( "@_ > /dev/null 2>&1" ) == 0;
 }
 
+sub qt1( $ ) {
+    1 while system( "@_ > /dev/null 2>&1" ) == 4;
+    $? == 0;
+}
+
 #
 # Determine which optional facilities are supported by iptables/netfilter
 #
@@ -1330,77 +1362,77 @@ sub determine_capabilities( $ ) {
     my $pid       = $$;
     my $sillyname = "fooX$pid";
 
-    $capabilities{NAT_ENABLED}    = qt( "$iptables -t nat -L -n" );
-    $capabilities{MANGLE_ENABLED} = qt( "$iptables -t mangle -L -n" );
+    $capabilities{NAT_ENABLED}    = qt1( "$iptables -t nat -L -n" );
+    $capabilities{MANGLE_ENABLED} = qt1( "$iptables -t mangle -L -n" );
 
-    qt( "$iptables -N $sillyname" );
+    qt1( "$iptables -N $sillyname" );
 
-    $capabilities{CONNTRACK_MATCH} = qt( "$iptables -A $sillyname -m conntrack --ctorigdst 192.168.1.1 -j ACCEPT" );
+    $capabilities{CONNTRACK_MATCH} = qt1( "$iptables -A $sillyname -m conntrack --ctorigdst 192.168.1.1 -j ACCEPT" );
     
-    if ( qt( "$iptables -A $sillyname -p tcp -m multiport --dports 21,22 -j ACCEPT" ) ) {
+    if ( qt1( "$iptables -A $sillyname -p tcp -m multiport --dports 21,22 -j ACCEPT" ) ) {
 	$capabilities{MULTIPORT}  = 1;
-	$capabilities{KLUDGEFREE} = qt( "$iptables -A $sillyname -p tcp -m multiport --sports 60 -m multiport --dports 99 -j ACCEPT" );
+	$capabilities{KLUDGEFREE} = qt1( "$iptables -A $sillyname -p tcp -m multiport --sports 60 -m multiport --dports 99 -j ACCEPT" );
     }
 
-    $capabilities{XMULTIPORT}      = qt( "$iptables -A $sillyname -p tcp -m multiport --dports 21:22 -j ACCEPT" );
-    $capabilities{POLICY_MATCH}    = qt( "$iptables -A $sillyname -m policy --pol ipsec --mode tunnel --dir in -j ACCEPT" );
+    $capabilities{XMULTIPORT}      = qt1( "$iptables -A $sillyname -p tcp -m multiport --dports 21:22 -j ACCEPT" );
+    $capabilities{POLICY_MATCH}    = qt1( "$iptables -A $sillyname -m policy --pol ipsec --mode tunnel --dir in -j ACCEPT" );
 
-    if ( qt( "$iptables -A $sillyname -m physdev --physdev-in eth0 -j ACCEPT" ) ) {
+    if ( qt1( "$iptables -A $sillyname -m physdev --physdev-in eth0 -j ACCEPT" ) ) {
 	$capabilities{PHYSDEV_MATCH}  = 1;
-	$capabilities{PHYSDEV_BRIDGE} = qt( "$iptables -A $sillyname -m physdev --physdev-is-bridged --physdev-in eth0 --physdev-out eth1 -j ACCEPT" );
+	$capabilities{PHYSDEV_BRIDGE} = qt1( "$iptables -A $sillyname -m physdev --physdev-is-bridged --physdev-in eth0 --physdev-out eth1 -j ACCEPT" );
 	unless ( $capabilities{KLUDGEFREE} ) {
-	    $capabilities{KLUDGEFREE} = qt( "$iptables -A $sillyname -m physdev --physdev-in eth0 -m physdev --physdev-out eth0 -j ACCEPT" );
+	    $capabilities{KLUDGEFREE} = qt1( "$iptables -A $sillyname -m physdev --physdev-in eth0 -m physdev --physdev-out eth0 -j ACCEPT" );
 	}
     }
 
-    if ( qt( "$iptables -A $sillyname -m iprange --src-range 192.168.1.5-192.168.1.124 -j ACCEPT" ) ) {
+    if ( qt1( "$iptables -A $sillyname -m iprange --src-range 192.168.1.5-192.168.1.124 -j ACCEPT" ) ) {
 	$capabilities{IPRANGE_MATCH} = 1;
 	unless ( $capabilities{KLUDGEFREE} ) {
-	    $capabilities{KLUDGEFREE} = qt( "$iptables -A $sillyname -m iprange --src-range 192.168.1.5-192.168.1.124 -m iprange --dst-range 192.168.1.5-192.168.1.124 -j ACCEPT" );
+	    $capabilities{KLUDGEFREE} = qt1( "$iptables -A $sillyname -m iprange --src-range 192.168.1.5-192.168.1.124 -m iprange --dst-range 192.168.1.5-192.168.1.124 -j ACCEPT" );
 	}
     }
 
-    $capabilities{RECENT_MATCH} = qt( "$iptables -A $sillyname -m recent --update -j ACCEPT" );
-    $capabilities{OWNER_MATCH}  = qt( "$iptables -A $sillyname -m owner --uid-owner 0 -j ACCEPT" );
+    $capabilities{RECENT_MATCH} = qt1( "$iptables -A $sillyname -m recent --update -j ACCEPT" );
+    $capabilities{OWNER_MATCH}  = qt1( "$iptables -A $sillyname -m owner --uid-owner 0 -j ACCEPT" );
 
-    if ( qt( "$iptables -A $sillyname -m connmark --mark 2  -j ACCEPT" )) {
+    if ( qt1( "$iptables -A $sillyname -m connmark --mark 2  -j ACCEPT" )) {
 	$capabilities{CONNMARK_MATCH}  = 1;
-	$capabilities{XCONNMARK_MATCH} = qt( "$iptables -A $sillyname -m connmark --mark 2/0xFF -j ACCEPT" );
+	$capabilities{XCONNMARK_MATCH} = qt1( "$iptables -A $sillyname -m connmark --mark 2/0xFF -j ACCEPT" );
     }
 
-    $capabilities{IPP2P_MATCH}     = qt( "$iptables -A $sillyname -p tcp -m ipp2p --ipp2p -j ACCEPT" );
-    $capabilities{LENGTH_MATCH}    = qt( "$iptables -A $sillyname -m length --length 10:20 -j ACCEPT" );
-    $capabilities{ENHANCED_REJECT} = qt( "$iptables -A $sillyname -j REJECT --reject-with icmp-host-prohibited" );
-    $capabilities{COMMENTS}        = qt( qq($iptables -A $sillyname -j ACCEPT -m comment --comment "This is a comment" ) );
+    $capabilities{IPP2P_MATCH}     = qt1( "$iptables -A $sillyname -p tcp -m ipp2p --ipp2p -j ACCEPT" );
+    $capabilities{LENGTH_MATCH}    = qt1( "$iptables -A $sillyname -m length --length 10:20 -j ACCEPT" );
+    $capabilities{ENHANCED_REJECT} = qt1( "$iptables -A $sillyname -j REJECT --reject-with icmp-host-prohibited" );
+    $capabilities{COMMENTS}        = qt1( qq($iptables -A $sillyname -j ACCEPT -m comment --comment "This is a comment" ) );
 
     if  ( $capabilities{MANGLE_ENABLED} ) {
-	qt( "$iptables -t mangle -N $sillyname" );
+	qt1( "$iptables -t mangle -N $sillyname" );
 
-	if ( qt( "$iptables -t mangle -A $sillyname -j MARK --set-mark 1" ) ) {
+	if ( qt1( "$iptables -t mangle -A $sillyname -j MARK --set-mark 1" ) ) {
 	    $capabilities{MARK}  = 1;
-	    $capabilities{XMARK} = qt( "$iptables -t mangle -A $sillyname -j MARK --and-mark 0xFF" );
+	    $capabilities{XMARK} = qt1( "$iptables -t mangle -A $sillyname -j MARK --and-mark 0xFF" );
 	}
 
-	if ( qt( "$iptables -t mangle -A $sillyname -j CONNMARK --save-mark" ) ) {
+	if ( qt1( "$iptables -t mangle -A $sillyname -j CONNMARK --save-mark" ) ) {
 	    $capabilities{CONNMARK}  = 1;
-	    $capabilities{XCONNMARK} = qt( "$iptables -t mangle -A $sillyname -j CONNMARK --save-mark --mask 0xFF" );
+	    $capabilities{XCONNMARK} = qt1( "$iptables -t mangle -A $sillyname -j CONNMARK --save-mark --mask 0xFF" );
 	}
 
-	$capabilities{CLASSIFY_TARGET} = qt( "$iptables -t mangle -A $sillyname -j CLASSIFY --set-class 1:1" );
-	qt( "$iptables -t mangle -F $sillyname" );
-	qt( "$iptables -t mangle -X $sillyname" );
+	$capabilities{CLASSIFY_TARGET} = qt1( "$iptables -t mangle -A $sillyname -j CLASSIFY --set-class 1:1" );
+	qt1( "$iptables -t mangle -F $sillyname" );
+	qt1( "$iptables -t mangle -X $sillyname" );
 
-	$capabilities{MANGLE_FORWARD} = qt( "$iptables -t mangle -L FORWARD -n" );
+	$capabilities{MANGLE_FORWARD} = qt1( "$iptables -t mangle -L FORWARD -n" );
     }
 
-    $capabilities{RAW_TABLE} = qt( "$iptables -t raw -L -n" );
+    $capabilities{RAW_TABLE} = qt1( "$iptables -t raw -L -n" );
 
     if ( which 'ipset' ) {
 	qt( "ipset -X $sillyname" );
 
 	if ( qt( "ipset -N $sillyname iphash" ) ) {
-	    if ( qt( "$iptables -A $sillyname -m set --set $sillyname src -j ACCEPT" ) ) {
-		qt( "$iptables -D $sillyname -m set --set $sillyname src -j ACCEPT" );
+	    if ( qt1( "$iptables -A $sillyname -m set --set $sillyname src -j ACCEPT" ) ) {
+		qt1( "$iptables -D $sillyname -m set --set $sillyname src -j ACCEPT" );
 		$capabilities{IPSET_MATCH} = 1;
 	    }
 
@@ -1408,14 +1440,14 @@ sub determine_capabilities( $ ) {
 	}
     }
 
-    $capabilities{USEPKTTYPE}      = qt( "$iptables -A $sillyname -m pkttype --pkt-type broadcast -j ACCEPT" );
-    $capabilities{ADDRTYPE}        = qt( "$iptables -A $sillyname -m addrtype --src-type BROADCAST -j ACCEPT" );
-    $capabilities{TCPMSS_MATCH}    = qt( "$iptables -A $sillyname -p tcp --tcp-flags SYN,RST SYN -m tcpmss --mss 1000:1500 -j ACCEPT" );
-    $capabilities{HASHLIMIT_MATCH} = qt( "$iptables -A $sillyname -m hashlimit --hashlimit 4 --hashlimit-burst 5 --hashlimit-name fooX1234 --hashlimit-mode dstip -j ACCEPT" );
-    $capabilities{NFQUEUE_TARGET}  = qt( "$iptables -A $sillyname -j NFQUEUE --queue-num 4" );
+    $capabilities{USEPKTTYPE}      = qt1( "$iptables -A $sillyname -m pkttype --pkt-type broadcast -j ACCEPT" );
+    $capabilities{ADDRTYPE}        = qt1( "$iptables -A $sillyname -m addrtype --src-type BROADCAST -j ACCEPT" );
+    $capabilities{TCPMSS_MATCH}    = qt1( "$iptables -A $sillyname -p tcp --tcp-flags SYN,RST SYN -m tcpmss --mss 1000:1500 -j ACCEPT" );
+    $capabilities{HASHLIMIT_MATCH} = qt1( "$iptables -A $sillyname -m hashlimit --hashlimit 4 --hashlimit-burst 5 --hashlimit-name fooX1234 --hashlimit-mode dstip -j ACCEPT" );
+    $capabilities{NFQUEUE_TARGET}  = qt1( "$iptables -A $sillyname -j NFQUEUE --queue-num 4" );
 
-    qt( "$iptables -F $sillyname" );
-    qt( "$iptables -X $sillyname" );
+    qt1( "$iptables -F $sillyname" );
+    qt1( "$iptables -X $sillyname" );
 
     $capabilities{CAPVERSION} = $globals{CAPVERSION};
 }
@@ -1674,6 +1706,7 @@ sub get_configuration( $ ) {
     default_yes_no 'EXPAND_POLICIES'            , '';
     default_yes_no 'KEEP_RT_TABLES'             , '';
     default_yes_no 'DELETE_THEN_ADD'            , 'Yes';
+    default_yes_no 'BROKEN_ROUTING'             , '';
     default_yes_no 'MULTICAST'                  , '';
     default_yes_no 'MARK_IN_FORWARD_CHAIN'      , '';
     
@@ -1939,7 +1972,9 @@ sub generate_aux_config() {
 
     create_temp_aux_config;
 
-    emit join ( '', "#\n# Shorewall auxiliary configuration file created by Shorewall-perl version ", $globals{VERSION}, ' - ' , localtime , "\n#" );
+    my $date = localtime;
+
+    emit "#\n# Shorewall auxiliary configuration file created by Shorewall-perl version $globals{VERSION} - $date\n#";
 
     for my $option qw(VERBOSITY LOGFILE LOGFORMAT IPTABLES PATH SHOREWALL_SHELL SUBSYSLOCK LOCKFILE RESTOREFILE SAVE_IPSETS) {
 	conditionally_add_option $option;
