@@ -74,7 +74,7 @@ sub initialize() {
     $sectioned = 0;
     $macro_nest_level = 0;
     $current_param = '';
-    @param_stack;
+    @param_stack = ();
 }
 
 INIT {
@@ -213,6 +213,8 @@ sub setup_rfc1918_filteration( $ ) {
     my $rfc1918ref   = new_standard_chain 'rfc1918';
     my $chainref     = $norfc1918ref;
 
+    require_capability 'CONNTRACK_MATCH', "The norfc1918 option" , 's';
+
     log_rule $config{RFC1918_LOG_LEVEL} , $rfc1918ref , 'DROP' , '';
 
     add_rule $rfc1918ref , '-j DROP';
@@ -356,7 +358,7 @@ sub process_criticalhosts() {
 	my @hosts;
 
 	for my $host ( split /,/, $hosts ) {
-	    validate_net $host, 1;
+	    validate_host $host, 1;
 	    push @hosts, "$interface:$host";
 	}
 
@@ -397,7 +399,7 @@ sub process_routestopped() {
 	my @hosts;
 
 	for my $host ( split /,/, $hosts ) {
-	    validate_net $host, 1;
+	    validate_host $host, 1;
 	    push @hosts, "$interface:$host";
 	}
 
@@ -442,7 +444,7 @@ sub process_routestopped() {
 	my $desti   = match_dest_dev $interface;
 
 	emit "\$IPTABLES -A INPUT $sourcei $source -j ACCEPT";
-	emit "\$IPTABLES -A OUTPUT $desti $dest -j ACCEPT"    if $config{ADMINISABSENTMINDED};
+	emit "\$IPTABLES -A OUTPUT $desti $dest -j ACCEPT" unless $config{ADMINISABSENTMINDED};
 
 	my $matched = 0;
 
@@ -548,11 +550,12 @@ sub add_common_rules() {
 	}
     }
 
-    add_rule $rejectref , '-p tcp -j REJECT --reject-with tcp-reset';
+    add_rule $rejectref , '-p 2 -j DROP';
+    add_rule $rejectref , '-p 6 -j REJECT --reject-with tcp-reset';
 
     if ( $capabilities{ENHANCED_REJECT} ) {
-	add_rule $rejectref , '-p udp -j REJECT';
-	add_rule $rejectref, '-p icmp -j REJECT --reject-with icmp-host-unreachable';
+	add_rule $rejectref , '-p 17 -j REJECT';
+	add_rule $rejectref, '-p 1 -j REJECT --reject-with icmp-host-unreachable';
 	add_rule $rejectref, '-j REJECT --reject-with icmp-host-prohibited';
     } else {
 	add_rule $rejectref , '-j REJECT';
@@ -848,11 +851,12 @@ sub process_macro ( $$$$$$$$$$$$$ ) {
 	fatal_error "Invalid Action ($mtarget) in macro" unless $actiontype & ( ACTION +  STANDARD + NATRULE + MACRO );
 
 	if ( $msource ) {
-	    if ( ( $msource eq '-' ) || ( $msource eq 'SOURCE' ) ) {
+	    if ( $msource eq '-' ) {
 		$msource = $source || '';
-	    } elsif ( $msource eq 'DEST' ) {
-		$msource = $dest || '';
+	    } elsif ( $msource =~ s/^DEST:?// ) {
+		$msource = merge_macro_source_dest $msource, $dest; 
 	    } else {
+		$msource =~ s/^SOURCE:?//;
 		$msource = merge_macro_source_dest $msource, $source;
 	    }
 	} else {
@@ -860,11 +864,12 @@ sub process_macro ( $$$$$$$$$$$$$ ) {
 	}
 
 	if ( $mdest ) {
-	    if ( ( $mdest eq '-' ) || ( $mdest eq 'DEST' ) ) {
+	    if ( $mdest eq '-' ) {
 		$mdest = $dest || '';
-	    } elsif ( $mdest eq 'SOURCE' ) {
-		$mdest = $source || '';
+	    } elsif ( $mdest =~ s/^SOURCE:?// ) {
+		$mdest = merge_macro_source_dest $mdest , $source;
 	    } else {
+		$mdest =~ s/DEST:?//;
 		$mdest = merge_macro_source_dest $mdest, $dest;
 	    }
 	} else {
@@ -985,6 +990,7 @@ sub process_rule1 ( $$$$$$$$$$$ ) {
     my $destref;
 
     if ( $source =~ /^(.+?):(.*)/ ) {
+	fatal_error "Missing Source Qualifier ($source)" if $2 eq '';
 	$sourcezone = $1;
 	$source = $2;
     } else {
@@ -993,6 +999,7 @@ sub process_rule1 ( $$$$$$$$$$$ ) {
     }
 
     if ( $dest =~ /^(.+?):(.*)/ ) {
+	fatal_error "Missing DEST Qualifier ($dest)" if $2 eq '';
 	$destzone = $1;
 	$dest = $2;
     } else {
@@ -1131,7 +1138,11 @@ sub process_rule1 ( $$$$$$$$$$$ ) {
 	} else {
 	    fatal_error "A server must be specified in the DEST column in $action rules" if $server eq '';
 
-	    validate_address $server, 0;
+	    if ( $server =~ /^(.+)-(.+)$/ ) {
+		validate_range( $1, $2 );
+	    } else {
+		validate_address $server, 0;
+	    }
 
 	    if ( $action eq 'SAME' ) {
 		fatal_error 'Port mapping not allowed in SAME rules' if $serverport;
@@ -1508,7 +1519,7 @@ sub generate_matrix() {
 		my $rule = match_source_dev( $interface ) . match_source_net( $net ) . "-j RETURN";
 		add_rule $frwd_ref , $rule;
 		add_rule $in_ref   , $rule;
-		add_rule $out_ref  , $rule;
+		add_rule $out_ref  , match_dest_dev( $interface ) . match_dest_net( $net ) . "-j RETURN";
 	    }
 	}
 
@@ -1517,8 +1528,7 @@ sub generate_matrix() {
 	    my $source_ref = ( $zoneref->{hosts}{ipsec4} ) || {};
 
 	    if ( $config{DYNAMIC_ZONES} ) {
-		no warnings;
-		create_zone_dyn_chain $zone, $frwd_ref if (%$source_ref || $type eq 'ipsec4' );
+		create_zone_dyn_chain $zone, $frwd_ref if ( scalar keys %$source_ref ) || ( $type eq 'ipsec4' );
 	    }
 
 	    for my $interface ( keys %$source_ref ) {
@@ -1560,7 +1570,7 @@ sub generate_matrix() {
 
 	if ( $config{DYNAMIC_ZONES} ) {
 	    push @rule_chains , [ firewall_zone , $zone , $chain1 ] if $chain1;
-	    push @rule_chains , [ $zone , firewall_zone , $chain2 ];
+	    push @rule_chains , [ $zone , firewall_zone , $chain2 ] if $chain2;
 	}
 
 	#
@@ -1730,6 +1740,7 @@ sub generate_matrix() {
 		    for my $interface ( keys %$typeref ) {
 			my $arrayref = $typeref->{$interface};
 			for my $hostref ( @$arrayref ) {
+			    next if $hostref->{options}{sourceonly};
 			    if ( $zone ne $zone1 || $num_ifaces > 1 || $hostref->{options}{routeback} ) {
 				my $ipsec_out_match = match_ipsec_out $zone1 , $hostref;
 				for my $net ( @{$hostref->{hosts}} ) {
@@ -1751,6 +1762,7 @@ sub generate_matrix() {
 				    for my $interface1 ( keys %$type1ref ) {
 					my $array1ref = $type1ref->{$interface1};
 					for my $host1ref ( @$array1ref ) {
+					    next if $host1ref->{options}{sourceonly};
 					    my $ipsec_out_match = match_ipsec_out $zone1 , $host1ref;
 					    for my $net1 ( @{$host1ref->{hosts}} ) {
 						unless ( $interface eq $interface1 && $net eq $net1 && ! $host1ref->{options}{routeback} ) {
