@@ -68,22 +68,9 @@ our %tcs = ( T => { chain  => 'tcpost',
 		    fw       => 0
 		    } ,
 	    CF => { chain    => 'tcfor' ,
+		    connmark => 1 ,
 		    fw       => 0 ,
-		    connmark => 1 ,
 		    } ,
-	    T  => { chain    => 'tcpost' ,
-		    connmark => 0 ,
-		    fw       => 0
-		    } ,
-	    CT => { chain    => 'tcpost' ,
-		    target => 'CONNMARK --set-mark' ,
-		    connmark => 1 ,
-		    fw       => 0
-		    } ,
-	    C  => { target => 'CONNMARK --set-mark' ,
-		    connmark => 1 ,
-		    fw       => 0
-		    }
 	    );
 
 use constant { NOMARK    => 0 ,
@@ -94,17 +81,20 @@ use constant { NOMARK    => 0 ,
 our @tccmd = ( { match     => sub ( $ ) { $_[0] eq 'SAVE' } ,
 		 target    => 'CONNMARK --save-mark --mask' ,
 		 mark      => SMALLMARK ,
-		 mask      => '0xFF'
+		 mask      => '0xFF' ,
+		 connmark  => 1
 	       } ,
 	      { match     => sub ( $ ) { $_[0] eq 'RESTORE' },
 		target    => 'CONNMARK --restore-mark --mask' ,
 		mark      => SMALLMARK ,
-		mask      => '0xFF'
+		mask      => '0xFF' ,
+		connmark  => 1
 		} ,
 	      { match     => sub ( $ ) { $_[0] eq 'CONTINUE' },
 		target    => 'RETURN' ,
 		mark      => NOMARK ,
-		mask      => ''
+		mask      => '' ,
+		connmark  => 0
 		} ,
 	      { match     => sub ( $ ) { $_[0] =~ '\|.*'} ,
 		target    => 'MARK --or-mark' ,
@@ -113,7 +103,8 @@ our @tccmd = ( { match     => sub ( $ ) { $_[0] eq 'SAVE' } ,
 	      { match     => sub ( $ ) { $_[0] =~ '&.*' },
 		target    => 'MARK --and-mark ' ,
 		mark      => HIGHMARK ,
-		mask      => ''
+		mask      => '' ,
+		connmark  => 0
 		}
 	      );
 
@@ -152,6 +143,10 @@ our %tcclasses;
 
 our $prefix;
 
+our %restrictions = ( tcpre      => PREROUTE_RESTRICT ,
+		      tcpost     => POSTROUTE_RESTRICT ,
+		      tcfor      => NO_RESTRICT ,
+		      tcout      => OUTPUT_RESTRICT );
 #
 # Initialize globals -- we take this novel approach to globals initialization to allow
 #                       the compiler to run multiple times in the same process. The
@@ -176,13 +171,11 @@ INIT {
 }
 
 sub process_tc_rule( $$$$$$$$$$ ) {
-    my ( $mark, $source, $dest, $proto, $ports, $sports, $user, $testval, $length, $tos ) = @_;
+    my ( $originalmark, $source, $dest, $proto, $ports, $sports, $user, $testval, $length, $tos ) = @_;
 
-    my $original_mark = $mark;
+    my ( $mark, $designator, $remainder ) = split( /:/, $originalmark, 3 );
 
-    ( $mark, my ( $designator, $remainder ) ) = split( /:/, $mark, 3 );
-
-    fatal_error "Invalid MARK" if defined $remainder;
+    fatal_error "Invalid MARK ($originalmark)" if defined $remainder || ! defined $mark;
 
     my $chain  = $globals{MARKING_CHAIN};
     my $target = 'MARK --set-mark';
@@ -213,20 +206,22 @@ sub process_tc_rule( $$$$$$$$$$ ) {
 	    $target   = $tcsref->{target} if $tcsref->{target};
 	    $mark     = "$mark/0xFF"      if $connmark = $tcsref->{connmark};
 
+	    require_capability ('CONNMARK' , "CONNMARK Rules", '' ) if $connmark;
+
 	} else {
-	    fatal_error "Invalid MARK ($original_mark)"   unless $mark =~ /^([0-9]+|0x[0-9a-f]+)$/ and $designator =~ /^([0-9]+|0x[0-9a-f]+)$/;
+	    fatal_error "Invalid MARK ($originalmark)"   unless $mark =~ /^([0-9]+|0x[0-9a-f]+)$/ and $designator =~ /^([0-9]+|0x[0-9a-f]+)$/;
 
 	    if ( $config{TC_ENABLED} eq 'Internal' ) {
-		fatal_error "Unknown Class ($original_mark)}" unless ( $device = $classids{$original_mark} );
+		fatal_error "Unknown Class ($originalmark)}" unless ( $device = $classids{$originalmark} );
 	    }
 
 	    $chain   = 'tcpost';
 	    $classid = 1;
-	    $mark    = $original_mark;
+	    $mark    = $originalmark;
 	    $target  = 'CLASSIFY --set-class';
 	}
     }
-
+        
     my $mask = 0xffff;
 
     my ($cmd, $rest) = split( '/', $mark, 2 );
@@ -237,6 +232,8 @@ sub process_tc_rule( $$$$$$$$$$ ) {
 	    for my $tccmd ( @tccmd ) {
 		if ( $tccmd->{match}($cmd) ) {
 		    fatal_error "$mark not valid with :C[FPT]" if $connmark;
+		    
+		    require_capability ('CONNMARK' , "SAVE/RESTORE Rules", '' ) if $tccmd->{connmark};
 
 		    $target      = "$tccmd->{target} ";
 		    my $marktype = $tccmd->{mark};
@@ -248,7 +245,7 @@ sub process_tc_rule( $$$$$$$$$$ ) {
 		    }
 
 		    if ( $rest ) {
-			fatal_error "Invalid MARK ($original_mark)" if $marktype == NOMARK;
+			fatal_error "Invalid MARK ($originalmark)" if $marktype == NOMARK;
 
 			$mark = $rest if $tccmd->{mask};
 
@@ -273,7 +270,7 @@ sub process_tc_rule( $$$$$$$$$$ ) {
     }
 
     if ( ( my $result = expand_rule( ensure_chain( 'mangle' , $chain ) ,
-				     NO_RESTRICT ,
+				     $restrictions{$chain} ,
 				     do_proto( $proto, $ports, $sports) . do_user( $user ) . do_test( $testval, $mask ) . do_tos( $tos ) ,
 				     $source ,
 				     $dest ,
@@ -286,7 +283,7 @@ sub process_tc_rule( $$$$$$$$$$ ) {
 	#
 	# expand_rule() returns destination device if any
 	#
-	fatal_error "Class Id $original_mark is not associated with device $result" if $device ne $result;
+	fatal_error "Class Id $originalmark is not associated with device $result" if $device ne $result;
     }
 
     progress_message "   TC Rule \"$currentline\" $done";
@@ -300,7 +297,7 @@ sub rate_to_kbit( $ ) {
     return $1 * 1000   if $rate =~ /^(\d+)mbit$/i;
     return $1 * 8000   if $rate =~ /^(\d+)mbps$/i;
     return $1 * 8      if $rate =~ /^(\d+)kbps$/i;
-    return $rate / 125 if $rate =~ /^\d+$/;
+    return int($1/125) if $rate =~ /^(\d+)(bps)?$/;
     fatal_error "Invalid Rate ($rate)";
 }
 
@@ -360,7 +357,7 @@ sub validate_tc_class( $$$$$$ ) {
     $tcclasses{$device} = {} unless $tcclasses{$device};
     my $tcref = $tcclasses{$device};
 
-    fatal_error "Invalid Mark ($mark)" unless $mark =~ /^([0-9]+|0x[0-9a-f]+)$/ && numeric_value( $mark ) <= 0xff;
+    fatal_error "Invalid Mark ($mark)" unless $mark =~ /^([0-9]+|0x[0-9a-fA-F]+)$/ && numeric_value( $mark ) <= 0xff;
 
     my $markval = numeric_value( $mark );
     fatal_error "Duplicate Mark ($mark)" if $tcref->{$markval};
